// Copyright (C) 2012-2014 Damir Korencic, Strahil Ristov
//
// This file is part of lzlex v0.1
//
// lzlex is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// lzlex is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lzlex.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FILEREADER_H
#define	FILEREADER_H

#include <fstream>
#include <vector>
#include <cstring>
#include <cstddef>
#include <iostream>

#include "node_array/types/symbol.h"
#include "dictionary/util/WordList.h"
#include "util/utils.h"

using namespace std;

//TODO Make WFR return char strings and let client convert them to TSymbols

/** Reads words from the file, one per line,
 * and stores them in memory for indexed access. */
template <typename TSymbol> class WordFileReader {
public:

    WordFileReader();
    WordFileReader(string fileName);
    virtual ~WordFileReader();

    void setFileName(string name);
    // read words (separated by whitespace) from file
    void readWords();

    //TODO remove functionality of word storage and retrieval, that
    // is the role of WordList
    size_t getNumberOfWords();
    TSymbol const * getWord(size_t i);

    WordList<TSymbol>* getWords();

    static TSymbol* readLine(istream& stream);

private:

    string fileName;
    FILE * file;
    vector<TSymbol *> words;

    void clearWords();


};

template <typename TSymbol>
WordFileReader<TSymbol>::WordFileReader(): fileName("") { }

template <typename TSymbol>
WordFileReader<TSymbol>::WordFileReader(string name) : fileName(name) { }

template <typename TSymbol>
WordFileReader<TSymbol>::~WordFileReader() {
    clearWords();
}

/** Set a new name for the word file and clear word that were read
 * from the previous file. */
template <typename TSymbol>
void WordFileReader<TSymbol>::setFileName(string name) {
    fileName = name;
    clearWords();
}

template <typename TSymbol>
size_t WordFileReader<TSymbol>::getNumberOfWords() {
    return words.size();
}

template <typename TSymbol>
TSymbol const * WordFileReader<TSymbol>::getWord(size_t i) {
    if (i >= words.size()) return NULL;
    return words[i];
}

/** Return string that correspond to lines in the file. Empty lines
 * are ignored. */
template <typename TSymbol>
WordList<TSymbol>* WordFileReader<TSymbol>::getWords() {
    ifstream stream(fileName.c_str());
    WordList<TSymbol>* words = new WordList<TSymbol>();

    while(stream.good()) {
        TSymbol* line = readLine(stream);
        if (wordLength(line) > 0)
            words->addWord(line);

        delete [] line;
    }

    return words;
}

/** Read a line of characters from stream and convert to string of TSymbols.
 * Both CR+LF and LF are considered end of line delimiters. */
template <typename TSymbol>
TSymbol* WordFileReader<TSymbol>::readLine(istream& stream) {
    const char CR = 13, LF = 10;
    bool carriageReturn = false;
    vector<char> buffer;

    while (stream.good()) {
        char ch = (char)stream.get();
        if (stream.gcount() == 0) break;
        
        if (ch == LF) { // end of line
            // remove carriage return if it preceedes line feed
            if (carriageReturn) buffer.pop_back();
            break;
        }
        else {
            if (ch == CR) carriageReturn = true;
            else carriageReturn = false;

            buffer.push_back(ch);
        }
    }

    TSymbol* line = new TSymbol[buffer.size()+1];
    size_t i;
    for (i = 0; i < buffer.size(); ++i) line[i] = (TSymbol)buffer[i];
    line[i] = zeroSymbol<TSymbol>();

    return line;
}

template <typename TSymbol>
void WordFileReader<TSymbol>::readWords() {
    ifstream stream(fileName.c_str());
    clearWords();
    const int buffLen = 10000;
    char buffer[buffLen];

    while(stream.good()) {
        stream >> ws;
        if (stream.good() == false) break;
        stream.getline(buffer, buffLen);

        const int size = strlen(buffer);
        TSymbol* word = new TSymbol[size+1];

        for (int i = 0; i < size; ++i)
            word[i] = (TSymbol)buffer[i];

        word[size] = zeroSymbol<TSymbol>();

        words.push_back(word);
    }
}

template <typename TSymbol>
void WordFileReader<TSymbol>::clearWords() {
    typename vector<TSymbol *>::iterator it;
    for (it = words.begin(); it != words.end(); ++it )
        delete [] (*it);
    
    words.clear();
}

#endif	/* FILEREADER_H */


