// Copyright (C) 2012-2014 Damir Korencic, Strahil Ristov
//
// This file is part of lzlex v0.1
//
// lzlex is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// lzlex is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lzlex.  If not, see <http://www.gnu.org/licenses/>.

#ifndef FACTORY_H
#define	FACTORY_H

#include <fstream>
#include <cstddef>

#include "dictionary/util/WordList.h"
#include "dictionary/char_trie/Trie.h"
#include "dictionary/lz_trie/LzTrie.h"
#include "dictionary/char_trie/WordIndexer.h"
#include "compress/lz_compressor/LzCompressor.h"
#include "node_array/compact_array/CompactArray.h"
#include "node_array/compact_array/CompactArrayCreator.h"
#include "serialization/array/CompactArraySerializer.h"
#include "compress/sa_compressor/LCPTreeCompressor.hpp"

using namespace std;

template <typename TNodeArray>
Trie<TNodeArray>* getTrie(WordList<typename TNodeArray::Symbol>& words, bool enumerated = false) {
    Trie<TNodeArray>* trie = new Trie<TNodeArray>(0, enumerated);

    for (size_t i = 0; i < words.numberOfWords(); ++i)
        trie->insertWord(words[i]);

    return trie;
}

/** Construct and return node array from a list of words. */
template <typename TNodeArray>
TNodeArray* getArray(WordList<typename TNodeArray::Symbol>& words) {
    Trie<TNodeArray>* trie = getTrie<TNodeArray>(words);
    TNodeArray* array = &(trie->exportNodeArray());
    delete trie;
    return array;
}

/** Construct and return enumerated node array from a list of words. */
template <typename TNodeArray>
TNodeArray* getEnumArray(WordList<typename TNodeArray::Symbol>& words) {
    Trie<TNodeArray>* trie = getTrie<TNodeArray>(words, true);
    TNodeArray* array = &(trie->exportNodeArray());
    delete trie;

    WordIndexer<TNodeArray, typename TNodeArray::Index> windex(*array);
    windex.enumerate();    
    
    return array;
}

template <typename TNodeArray>
TNodeArray* getLzArray(WordList<typename TNodeArray::Symbol>& words, bool enumerated = false) {
    TNodeArray* array;
    if (enumerated) array = getEnumArray<TNodeArray>(words);
    else array = getArray<TNodeArray>(words);
    LzCompressor<TNodeArray> compressor;
    //LCPTreeCompressor<TNodeArray> compressor;
    compressor.compressArray(*array);
    return array;
}

template <typename TNodeArray>
TNodeArray* getLzArrayLCT(WordList<typename TNodeArray::Symbol>& words, bool enumerated = false) {
    TNodeArray* array;
    if (enumerated) array = getEnumArray<TNodeArray>(words);
    else array = getArray<TNodeArray>(words);
    //LzCompressor<TNodeArray> compressor;
    LCPTreeCompressor<TNodeArray> compressor;
    compressor.compressArray(*array);
    return array;
}

template <typename TNodeArray>
LzTrie<TNodeArray>* getLzTrie(WordList<typename TNodeArray::Symbol>& words, bool enumerated = false) {
    LzTrie<TNodeArray>* trie =
            new LzTrie<TNodeArray>(*getLzArray<TNodeArray>(words, enumerated));
    return trie;
}

/** Get Compact array constructed from lz-compressed TNodeArray */
template <typename TNodeArray>
CompactArray<typename TNodeArray::Symbol, typename TNodeArray::Index> *
getCompactedLzArray(WordList<typename TNodeArray::Symbol>& words, bool enumerated = false) {
    TNodeArray* nodes = getLzArray<TNodeArray>(words, enumerated);

    CompactArrayCreator<TNodeArray> creator(*nodes);
    CompactArray<typename TNodeArray::Symbol, typename TNodeArray::Index> *carray =
            creator.createCompactArray();

    delete nodes;
    return carray;
}

template <typename TSymbol, typename TIndex>
LzTrie<CompactArray<TSymbol, TIndex> > * getLzTrieFromCompressedFile(string file) {
    CompactArraySerializer<TSymbol, TIndex> ser;
    fstream stream(file.c_str());
    if (stream.good()) {
        CompactArray<TSymbol, TIndex>* array = ser.arrayFromStream(stream);

        LzTrie<CompactArray<TSymbol, TIndex> >* lzTrie =
                new LzTrie<CompactArray<TSymbol, TIndex> >(*array);

        return lzTrie;
    }
    else return NULL;
}

#endif	/* FACTORY_H */
