// Copyright (C) 2012-2014 Damir Korencic, Strahil Ristov
//
// This file is part of lzlex v0.1
//
// lzlex is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// lzlex is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with lzlex.  If not, see <http://www.gnu.org/licenses/>.

#include <cstdlib>
#include <iostream>
#include <string>
#include <map>

using namespace std;

#include "util/WordFileReader.h"
#include "util/factory.h"
#include "util/regex.h"
#include "dictionary/util/WordList.h"
#include "node_array/vector_array/VectorArray.h"
#include "node_array/compact_array/CompactArray.h"
#include "node_array/compact_array/CompactArrayCreator.h"
#include "serialization/array/CompactArraySerializer.h"
#include "compress/stats/StatCreator.hpp"


using namespace std;

map<string, string> params;

typedef int TIndex;
typedef unsigned char TSymbol;
typedef VectorArray<TSymbol, TIndex> TNodeArray;
typedef CompactArray<TSymbol, TIndex> TCompactArray;

static void printUsage();
static void compress();
static void doCompress(string input_file, string output_file);
static void search();
static void intersect();
static void batchRun();
static void createParameterMap(int argc, char** argv);
static TSymbol* stringToTSymbolString(string& str);

/** Check number of arguments, then invoke the main command
 * depending on the first argument. */
int main(int argc, char** argv) {    
    if (argc < 2) {
        printUsage();
    }
    else {
        string command = argv[1];
        createParameterMap(argc-2, argv+2);
        if (command == "c") {
            compress();
        }
        else if (command == "s") {
            search();
        }
        else if (command == "i") {
            intersect();
        }
        else if (command == "b") {
            batchRun();
        }
        else {
            printUsage();
        }
    }

    return 0;
}

void createParameterMap(int argc, char** argv) {
    for (int i = 0; i < argc; ++i) {
        string arg = argv[i];
        // arguments for wich parameters are expected
        if (arg == "-i" || arg == "-l" || arg == "-s" || 
                arg == "-a" || arg == "-f" || arg == "-o") {
            if (i < argc - 1) {
                string val = argv[i+1];
                if (val.size() > 0 && val[0] != '-') params[arg] = val;
            }
        } else {
            // flag arguments
            if (arg == "-cmm") params[arg] = "rjecnik.cmm";
            else if (arg == "-st") params[arg] = "true";
            else if (arg == "-t") params[arg] = "true";
            else if (arg == "-e") params[arg] = "true";
        }
    }
    // print parameters
//    for (map<string, string>::iterator it = params.begin(); it != params.end(); ++it) {
//        cout<<(*it).first<<" "<<(*it).second<<endl;
//    }
}

void printUsage() {
    cout<<endl<<"LzLex lexicon utility"<<endl<<endl
        <<"compress: lzlex c -i input_file -l lexicon_file"<<endl
        //<<"compress: lzlex c -i input_file -l lexicon_file [-cmm]"<<endl
        <<"   Input file must be sorted lexicographically."<<endl<<endl            
        //<<"   Use -cmm flag to store compressed trie in .cmm file"<<endl<<endl
        <<"search: lzlex s -l lexicon_file [-s search_term]"<<endl
        <<"   If search term is ommited, read search queries from standard input."<<endl
        <<"   If the queries are read from stdin, '^' terminates the input."<<endl
        <<"   Search term is either a word or word*, * must be the last char and can be the only char. "<<endl
        <<"   Do not forget to escape * with a '\\' if necessary!"<<endl<<endl
        <<"intersect: lzlex i -l lexicon_file -i input_file "<<endl
        <<"   For each word in the input file print the word on stdout if it is in the dictionary."<<endl<<endl;
//        <<"batch: lzt b -a algorithms [-d dict_folder] [-f dict_files] [-o out_folder] [-st][-t][-e]"<<endl
//        <<"   Do a batch run, ie run each of the specified algorithms on each od the dict_files" << endl
//        <<"   inside dict_folder, or all files in it if dict_files is ommited. " << endl
//        <<"   For -st, write statistics, for -t, test, ie compare words in compressed dictionary" << endl
//        <<"   with original  files, and for -e end run when first (comparison) error occurs. " << endl << endl;
}

void printWordList(string query, LzTrie<CompactArray<TSymbol, TIndex> >* lzTrie) {
    // convery string of chars to string of TSymbols
    TSymbol *queryTS = stringToTSymbolString(query);

    if (query.find('*') != string::npos) { // list words
        TSymbol prefix[1000];      
        bool syntaxOk = getPrefixBeforeStar<TSymbol>(queryTS, prefix);
        if (syntaxOk) {            
            WordList<TSymbol>* list = lzTrie->getWordsByPrefix(prefix);
            wordListToStream<TSymbol>(list, cout);
            //cout<<"number of words: "<<list->numberOfWords()<<endl;
            delete list;
        }
        else {
            cout<<"* symbol must be the last symbol of the input."<<endl;
        }
    }
    else { // print a single word
        bool contains = lzTrie->containsWord(queryTS);
        if (contains) cout<<"in the dictionary"<<endl;
        else cout<<"NOT in the dictionary"<<endl;
    }

    delete [] queryTS;
}

void search() {
    if (params.count("-l") == 0) {
        cout<<"Please specify the lexicon file." << endl;
        printUsage();
        return;
    }
    bool stdInput = (params.count("-s") == 0);

    LzTrie<CompactArray<TSymbol, TIndex> >* lzTrie = 
            getLzTrieFromCompressedFile<TSymbol, TIndex>(params["-l"]);

    if (stdInput) {
        string input;
        while (true) {
            cin >> input;
            if (input == "^") break;
            printWordList(input, lzTrie);
        }
    }
    else {
        // replace # by * in the query string
//        string query = params["-s"];
//        for (int i = 0; i < query.size(); ++i)
//            if (query[i] == '#') query[i] = '*';

        printWordList(params["-s"], lzTrie);
    }

    delete lzTrie;
}

/** Executes compression of a list of words from input file
 * and writes compressed dictionary to the output file. */
void compress() {
    if (params.count("-i") == 0) {
        cout<<"Please specify the input file." << endl;
        printUsage();
        return;
    }
    if (params.count("-l") == 0) {
        params["-l"] = params["-i"]+".lzl";
    }

    doCompress(params["-i"], params["-l"]);
}

void writeWordsToFile(const WordList<TSymbol>* words, string fileName, bool linuxEow = false) {
    ofstream file; file.open(fileName.c_str(), ios::out);
    char buff[10000];

    for (size_t i = 0; i < words->numberOfWords(); ++i) {
        size_t len;
        const TSymbol * word = (*words)[i];
    
        for (len = 0; word[len] != zeroSymbol<TSymbol>(); ++len);
        for (size_t j = 0; j < len; ++j) buff[j] = (char)word[j];

        if (linuxEow) {
            buff[len] = 10;
            file.write(buff, len+1);
        }
        else {
            buff[len] = 13; buff[len+1] = 10;
            file.write(buff, len+2);
        }
    }

    file.close();
}

void doCompress(string inputFile, string outputFile) {
    cout<<"compressing "<<inputFile<<" to "<<outputFile<<" ..."<<endl;
    // create compressed array
    WordFileReader<TSymbol> reader(inputFile);
    WordList<TSymbol>* words = reader.getWords();
    words->sort();
    TNodeArray* array = getLzArrayLCT<TNodeArray>(*words);
    delete words;

    // create compact array
    CompactArrayCreator<TNodeArray> compacter(*array);
    CompactArray<TSymbol, TIndex>* carray =  compacter.createCompactArray();
    delete array;
    // serialize to file
    ofstream output(outputFile.c_str());
    CompactArraySerializer<TSymbol, TIndex> serializer(carray);
    serializer.arrayToStream(output);
    output.close();
    delete carray;

    cout<<" compression sucessful"<<endl;
}

TSymbol* stringToTSymbolString(string& str) {
    TSymbol *tss = new TSymbol[str.size()+1];
    int i;
    for (i = 0; i < str.size(); ++i) tss[i] = (TSymbol)str[i];
    tss[i] = zeroSymbol<TSymbol>();
    return tss;
}

void intersect() {
    if (params.count("-i") == 0) {
        cout<<"Please specify the input file." << endl;
        printUsage();
        return;
    }
    if (params.count("-l") == 0) {
        cout<<"Please specify the lexicon file." << endl;
        printUsage();
        return;
    }
    LzTrie<CompactArray<TSymbol, TIndex> >* lzTrie =
            getLzTrieFromCompressedFile<TSymbol, TIndex>(params["-l"]);

    fstream file(params["-i"].c_str());
    while (file.good()) {
        string word;
        file >> ws;
        file >> word;
        TSymbol *tsWord = stringToTSymbolString(word);
        if (lzTrie->containsWord(tsWord)) cout<<word<<endl;
        delete [] tsWord;
    }

    delete lzTrie;
}

/** Run specified set of compression algorithms on a specified set of
 * dictionaries.  */
void batchRun() {
  //lzt b -a algorithms [-l lexicon_folder] [-f lexicon_files] [-o output_folder] [-st][-t][-e]

    if (params.count("-a") == 0) {
        cout << "Please specify algorithms." << endl;
        printUsage();
        return;
    }
    if (params.count("-l") == 0 && params.count("-f") == 0) {
        cout << "Please specifiy folder and/or files." << endl;
        printUsage();
        return;
    }

    bool stats = (params["-st"] == "true");
    bool test = (params["-t"] == "true");
    bool err = (params["-e"] == "true");

    StatCreator<TNodeArray> statCreator(stats, test, err);
    
    string dictFolder = (params.count("-l") == 0) ? "" : params["-l"];
    string dictFiles = (params.count("-f") == 0) ? "" : params["-f"];
    string outFolder = (params.count("-o") == 0) ? "" : params["-o"];
    string algos = params["-a"];

    statCreator.makeStats(algos, dictFolder, dictFiles, outFolder);
}
